# -*- coding: utf-8 -*-
import logging
import time
from typing import Dict, Hashable, List, Optional, Union

from pipgrip.libs.mixology.constraint import Constraint
from pipgrip.libs.mixology.failure import SolverFailure
from pipgrip.libs.mixology.incompatibility import Incompatibility
from pipgrip.libs.mixology.incompatibility_cause import (
    ConflictCause,
    NoVersionsCause,
    RootCause,
)
from pipgrip.libs.mixology.package_source import PackageSource
from pipgrip.libs.mixology.partial_solution import PartialSolution
from pipgrip.libs.mixology.range import Range
from pipgrip.libs.mixology.result import SolverResult
from pipgrip.libs.mixology.set_relation import SetRelation
from pipgrip.libs.mixology.term import Term

logger = logging.getLogger(__name__)

_conflict = object()


class VersionSolver:
    """
    The version solver that finds a set of package versions that satisfy the
    root package's dependencies.

    See https://github.com/dart-lang/pub/tree/master/doc/solver.md for details
    on how this solver works.
    """

    def __init__(
        self, source  # type: PackageSource
    ):
        self._source = source

        self._incompatibilities = {}  # type: Dict[Hashable, List[Incompatibility]]
        self._solution = PartialSolution()

    @property
    def solution(self):  # type: () -> PartialSolution
        return self._solution

    def is_solved(self):  # type: () -> bool
        return not self._solution.unsatisfied

    def solve(self):  # type: () -> SolverResult
        """
        Finds a set of dependencies that match the root package's constraints,
        or raises an error if no such set is available.
        """
        start = time.time()

        self._add_incompatibility(
            Incompatibility(
                [Term(Constraint(self._source.root, Range()), False)], RootCause()
            )
        )
        self._propagate(self._source.root)

        packages_tried = 0
        max_tries = -1
        while True:
            if packages_tried == max_tries:
                raise SolverFailure("Stopping, {} packages tried.".format(max_tries))
            if not self._run():
                break

            packages_tried += 1

        logger.info("Version solving took {:.3f} seconds.".format(time.time() - start))
        logger.info("Tried {} solutions.".format(self._solution.attempted_solutions))

        return SolverResult(
            self._solution.decisions, self._solution.attempted_solutions
        )

    def _run(self):  # type: () -> bool
        if self.is_solved():
            return False

        next_package = self._choose_package_version()
        self._propagate(next_package)

        if self.is_solved():
            return False

        return True

    def _propagate(self, package):  # type: (Hashable) -> None
        """
        Performs unit propagation on incompatibilities transitively
        related to package to derive new assignments for _solution.
        """
        changed = set()
        changed.add(package)

        while changed:
            package = changed.pop()

            # Iterate in reverse because conflict resolution tends to produce more
            # general incompatibilities as time goes on. If we look at those first,
            # we can derive stronger assignments sooner and more eagerly find
            # conflicts.
            for incompatibility in reversed(self._incompatibilities[package]):
                result = self._propagate_incompatibility(incompatibility)

                if result is _conflict:
                    # If the incompatibility is satisfied by the solution, we use
                    # _resolve_conflict() to determine the root cause of the conflict as a
                    # new incompatibility.
                    #
                    # It also backjumps to a point in the solution
                    # where that incompatibility will allow us to derive new assignments
                    # that avoid the conflict.
                    root_cause = self._resolve_conflict(incompatibility)

                    # Back jumping erases all the assignments we did at the previous
                    # decision level, so we clear [changed] and refill it with the
                    # newly-propagated assignment.
                    changed.clear()
                    changed.add(str(self._propagate_incompatibility(root_cause)))
                    break
                elif result is not None:
                    changed.add(result)

    def _propagate_incompatibility(
        self, incompatibility
    ):  # type: (Incompatibility) -> Union[str, _conflict, None]
        """
        If incompatibility is almost satisfied by _solution, adds the
        negation of the unsatisfied term to _solution.

        If incompatibility is satisfied by _solution, returns _conflict. If
        incompatibility is almost satisfied by _solution, returns the
        unsatisfied term's package name.

        Otherwise, returns None.
        """
        # The first entry in incompatibility.terms that's not yet satisfied by
        # _solution, if one exists. If we find more than one, _solution is
        # inconclusive for incompatibility and we can't deduce anything.
        unsatisfied = None

        for term in incompatibility.terms:
            relation = self._solution.relation(term)
            if relation == SetRelation.DISJOINT:
                # If term is already contradicted by _solution, then
                # incompatibility is contradicted as well and there's nothing new we
                # can deduce from it.
                return
            elif relation == SetRelation.OVERLAPPING:
                # If more than one term is inconclusive, we can't deduce anything about
                # incompatibility.
                if unsatisfied is not None and unsatisfied.package != term.package:
                    return

                # If exactly one term in incompatibility is inconclusive, then it's
                # almost satisfied and [term] is the unsatisfied term. We can add the
                # inverse of the term to _solution.
                unsatisfied = term

        # If *all* terms in incompatibility are satisfied by _solution, then
        # incompatibility is satisfied and we have a conflict.
        if unsatisfied is None:
            return _conflict

        logger.info("derived: {}".format(unsatisfied.inverse))

        self._solution.derive(
            unsatisfied.constraint, not unsatisfied.is_positive(), incompatibility
        )

        return unsatisfied.package

    def _resolve_conflict(
        self, incompatibility
    ):  # type: (Incompatibility) -> Incompatibility
        """
        Given an incompatibility that's satisfied by _solution,
        The `conflict resolution`_ constructs a new incompatibility that encapsulates the root
        cause of the conflict and backtracks _solution until the new
        incompatibility will allow _propagate() to deduce new assignments.

        Adds the new incompatibility to _incompatibilities and returns it.

        .. _conflict resolution: https://github.com/dart-lang/pub/tree/master/doc/solver.md#conflict-resolution
        """
        logger.info("conflict: {}".format(incompatibility))

        new_incompatibility = False
        while not incompatibility.is_failure():
            # The term in incompatibility.terms that was most recently satisfied by
            # _solution.
            most_recent_term = None

            # The earliest assignment in _solution such that incompatibility is
            # satisfied by _solution up to and including this assignment.
            most_recent_satisfier = None

            # The difference between most_recent_satisfier and most_recent_term;
            # that is, the versions that are allowed by most_recent_satisfier and not
            # by most_recent_term. This is None if most_recent_satisfier totally
            # satisfies most_recent_term.
            difference = None

            # The decision level of the earliest assignment in _solution *before*
            # most_recent_satisfier such that incompatibility is satisfied by
            # _solution up to and including this assignment plus
            # most_recent_satisfier.
            #
            # Decision level 1 is the level where the root package was selected. It's
            # safe to go back to decision level 0, but stopping at 1 tends to produce
            # better error messages, because references to the root package end up
            # closer to the final conclusion that no solution exists.
            previous_satisfier_level = 1

            for term in incompatibility.terms:
                satisfier = self._solution.satisfier(term)

                if most_recent_satisfier is None:
                    most_recent_term = term
                    most_recent_satisfier = satisfier
                elif most_recent_satisfier.index < satisfier.index:
                    previous_satisfier_level = max(
                        previous_satisfier_level, most_recent_satisfier.decision_level
                    )
                    most_recent_term = term
                    most_recent_satisfier = satisfier
                    difference = None
                else:
                    previous_satisfier_level = max(
                        previous_satisfier_level, satisfier.decision_level
                    )

                if most_recent_term == term:
                    # If most_recent_satisfier doesn't satisfy most_recent_term on its
                    # own, then the next-most-recent satisfier may be the one that
                    # satisfies the remainder.
                    difference = most_recent_satisfier.difference(most_recent_term)
                    if difference is not None:
                        previous_satisfier_level = max(
                            previous_satisfier_level,
                            self._solution.satisfier(difference.inverse).decision_level,
                        )

            # If most_recent_identifier is the only satisfier left at its decision
            # level, or if it has no cause (indicating that it's a decision rather
            # than a derivation), then incompatibility is the root cause. We then
            # backjump to previous_satisfier_level, where incompatibility is
            # guaranteed to allow _propagate to produce more assignments.
            if (
                previous_satisfier_level < most_recent_satisfier.decision_level
                or most_recent_satisfier.cause is None
            ):
                self._solution.backtrack(previous_satisfier_level)
                if new_incompatibility:
                    self._add_incompatibility(incompatibility)

                return incompatibility

            # Create a new incompatibility by combining incompatibility with the
            # incompatibility that caused most_recent_satisfier to be assigned. Doing
            # this iteratively constructs an incompatibility that's guaranteed to be
            # true (that is, we know for sure no solution will satisfy the
            # incompatibility) while also approximating the intuitive notion of the
            # "root cause" of the conflict.
            new_terms = []
            for term in incompatibility.terms:
                if term != most_recent_term:
                    new_terms.append(term)

            for term in most_recent_satisfier.cause.terms:
                if term.package != most_recent_satisfier.package:
                    new_terms.append(term)

            # The most_recent_satisfier may not satisfy most_recent_term on its own
            # if there are a collection of constraints on most_recent_term that
            # only satisfy it together. For example, if most_recent_term is
            # `foo ^1.0.0` and _solution contains `[foo >=1.0.0,
            # foo <2.0.0]`, then most_recent_satisfier will be `foo <2.0.0` even
            # though it doesn't totally satisfy `foo ^1.0.0`.
            #
            # In this case, we add `not (most_recent_satisfier \ most_recent_term)` to
            # the incompatibility as well, See the `algorithm documentation`_ for
            # details.
            #
            # .. _algorithm documentation: https://github.com/dart-lang/pub/tree/master/doc/solver.md#conflict-resolution
            if difference is not None:
                new_terms.append(difference.inverse)

            incompatibility = Incompatibility(
                new_terms, ConflictCause(incompatibility, most_recent_satisfier.cause)
            )
            new_incompatibility = True

            partially = "" if difference is None else " partially"
            bang = "!"
            logger.info(
                "{} {} is{} satisfied by {}".format(
                    bang, most_recent_term, partially, most_recent_satisfier
                )
            )
            logger.info(
                '{} which is caused by "{}"'.format(bang, most_recent_satisfier.cause)
            )
            logger.info("{} thus: {}".format(bang, incompatibility))

        raise SolverFailure(incompatibility)

    def _next_term_to_try(self):  # type: () -> Optional[Term]
        unsatisfied = self._solution.unsatisfied
        logger.debug("unsatisfied: %s", unsatisfied)
        if not unsatisfied:
            return

        # Prefer packages with as few remaining versions as possible,
        # so that if a conflict is necessary it's forced quickly.
        # at a tie, the package with least dependencies is chosen
        def _get_min(term):
            versions = self._source.versions_for(
                term.package, term.constraint.constraint
            )
            deps = (
                self._source.dependencies_for(term.package, versions[0])
                if versions
                else []
            )
            return len(versions), len(deps)

        if len(unsatisfied) == 1:
            term = unsatisfied[0]
        else:
            term = min(*unsatisfied, key=_get_min)

        return term

    def _choose_package_version(self):  # type: () -> Union[Hashable, None]
        """
        Tries to select a version of a required package.

        Returns the name of the package whose incompatibilities should be
        propagated by _propagate(), or None indicating that version solving is
        complete and a solution has been found.
        """
        term = self._next_term_to_try()
        if not term:
            return

        versions = self._source.versions_for(term.package, term.constraint.constraint)
        if not versions:
            # If there are no versions that satisfy the constraint,
            # add an incompatibility that indicates that.
            self._add_incompatibility(Incompatibility([term], NoVersionsCause()))

            return term.package

        version = versions[0]
        conflict = False
        for incompatibility in self._source.incompatibilities_for(
            term.package, version
        ):
            self._add_incompatibility(incompatibility)

            # If an incompatibility is already satisfied, then selecting version
            # would cause a conflict.
            #
            # We'll continue adding its dependencies, then go back to
            # unit propagation which will guide us to choose a better version.
            conflict = conflict or all(
                iterm.package == term.package
                and iterm.package.req.extras.issuperset(term.package.req.extras)
                or self._solution.satisfies(iterm)
                for iterm in incompatibility.terms
            )

        if not conflict:
            self._solution.decide(term.package, version)
            logger.info(
                "selecting {} ({})".format(term.package.req.extras_name, str(version))
            )

        return term.package

    def _add_incompatibility(self, incompatibility):  # type: (Incompatibility) -> None
        logger.info("fact: {}".format(incompatibility))

        for term in incompatibility.terms:
            if term.package not in self._incompatibilities:
                self._incompatibilities[term.package] = []

            if incompatibility in self._incompatibilities[term.package]:
                continue

            self._incompatibilities[term.package].append(incompatibility)



import io
import logging
import os
import re
import subprocess
import sys
from tempfile import NamedTemporaryFile

import pkg_resources
from click import echo as _echo
from packaging.markers import default_environment
from packaging.utils import canonicalize_name
from pkginfo import get_metadata

from pipgrip.compat import PIP_VERSION, urlparse

logger = logging.getLogger(__name__)


def read_requirements(path):
    re_comments = re.compile(r"(?:^|\s+)#")
    try:
        with io.open(path, mode="rt", encoding="utf-8") as fp:
            return list(
                filter(None, (re_comments.split(line, 1)[0].strip() for line in fp))
            )
    except IndexError:
        raise RuntimeError("{} is broken".format(path))


def parse_req(requirement, extras=None):
    if requirement == "_root_" or requirement == "." or requirement.startswith(".["):
        req = pkg_resources.Requirement.parse(
            requirement.replace(".", "rubbish", 1)
            if requirement.startswith(".[")
            else "rubbish"
        )
        if extras is not None:
            req.extras = extras
        req.key = "." if requirement.startswith(".[") else requirement
        full_str = req.__str__().replace(req.name, req.key)
        req.name = req.key
    else:
        req = pkg_resources.Requirement.parse(requirement)
        if extras is not None:
            req.extras = extras
        req.key = canonicalize_name(req.key)
        req.name = req.key
        full_str = req.__str__()  # .replace(req.name, req.key)

    def __str__():
        return full_str

    req.__str__ = __str__
    req.extras_name = (
        req.name + "[" + ",".join(sorted(req.extras)) + "]" if req.extras else req.name
    )
    req.extras = frozenset(req.extras)
    return req


def stream_bash_command(bash_command, echo=False):
    # https://gist.github.com/jaketame/3ed43d1c52e9abccd742b1792c449079
    # https://gist.github.com/bgreenlee/1402841
    logger.debug(bash_command)
    process = subprocess.Popen(
        bash_command,
        stdout=subprocess.PIPE,
        stderr=subprocess.STDOUT,
    )

    def check_io():
        output = ""
        while True:
            line = process.stdout.readline().decode("utf-8")
            if line:
                output += line
                if echo:
                    _echo(line[:-1])
            else:
                break
        return output

    # keep checking stdout/stderr until the child exits
    out = ""
    while process.poll() is None:
        out += check_io()

    return_code = process.wait()
    if return_code:
        raise subprocess.CalledProcessError(return_code, bash_command, output=out)

    return out


def _get_install_args(index_url, extra_index_url, pre, cache_dir, editable, user):
    args = [
        sys.executable,
        "-m",
        "pip",
        "install",
        "--cache-dir",
        cache_dir,
    ]
    if index_url is not None:
        args += [
            "--index-url",
            index_url,
            "--trusted-host",
            urlparse(index_url).hostname,
        ]
    if extra_index_url is not None:
        args += [
            "--extra-index-url",
            extra_index_url,
            "--trusted-host",
            urlparse(extra_index_url).hostname,
        ]
    if PIP_VERSION >= [10]:
        args.append("--progress-bar=off")
    if pre:
        args += ["--pre"]
    if editable:
        args += ["--editable"]
    if user:
        args += ["--user"]
    return args


def _get_wheel_args(index_url, extra_index_url, pre, cache_dir=None):
    args = [
        sys.executable,
        "-m",
        "pip",
        "wheel",
        "--no-deps",
        "--disable-pip-version-check",
    ]
    if pre:
        args += ["--pre"]
    if cache_dir is not None:
        args += [
            "--wheel-dir",
            cache_dir,
        ]
    if index_url is not None:
        args += [
            "--index-url",
            index_url,
            "--trusted-host",
            urlparse(index_url).hostname,
        ]
    if extra_index_url is not None:
        args += [
            "--extra-index-url",
            extra_index_url,
            "--trusted-host",
            urlparse(extra_index_url).hostname,
        ]
    if PIP_VERSION >= [10]:
        args.append("--progress-bar=off")
    return args


def install_packages(
    packages,
    index_url,
    extra_index_url,
    pre,
    cache_dir,
    editable,
    user,
    constraints=None,
):
    """Install a list of packages with pip."""
    args = (
        _get_install_args(index_url, extra_index_url, pre, cache_dir, editable, user)
        + packages
    )

    constraints_file = None
    if constraints is not None:
        logger.debug(constraints)
        with NamedTemporaryFile(delete=False, mode="w+") as fp:
            constraints_file = fp.name
            fp.write("\n".join(constraints) + "\n")
        args += ["--constraint", constraints_file]

    try:
        return stream_bash_command(args, echo=True)
    except subprocess.CalledProcessError as err:
        output = getattr(err, "output") or ""
        logger.error(output)
        raise
    finally:
        if constraints_file is not None:
            os.remove(constraints_file)


_available_versions_cache = {}


def _get_available_versions(package, index_url, extra_index_url, pre):
    cache_key = (package, pre)
    if cache_key in _available_versions_cache:
        return _available_versions_cache[cache_key]

    logger.debug("Finding possible versions for {}".format(package))
    args = _get_wheel_args(index_url, extra_index_url, pre) + [
        package + "==42.42.post424242"
    ]

    if [20, 3] <= PIP_VERSION < [21, 1]:
        # https://github.com/ddelange/pipgrip/issues/42
        args += ["--use-deprecated", "legacy-resolver"]

    try:
        out = stream_bash_command(args)
    except subprocess.CalledProcessError as err:
        # expected. we forced this by using a non-existing version number.
        out = getattr(err, "output") or ""
    else:
        logger.warning(out)
        raise RuntimeError("Unexpected success:" + " ".join(args))
    out = out.splitlines()
    for line in out[::-1]:
        if "Could not find a version that satisfies the requirement" in line:
            all_versions = line.split("from versions: ", 1)[1].rstrip(")").split(", ")
            if pre:
                return all_versions
            # filter out pre-releases
            logger.debug(
                str(
                    {
                        package: [
                            v for v in all_versions if not re.findall(r"[a-zA-Z]", v)
                        ]
                    }
                )
            )
            available_versions = [
                v for v in all_versions if not re.findall(r"[a-zA-Z]", v)
            ]
            _available_versions_cache[cache_key] = available_versions
            return available_versions
    raise RuntimeError("Failed to get available versions for {}".format(package))


def _download_wheel(package, index_url, extra_index_url, pre, cache_dir):
    """Download/build wheel for package and return its filename."""
    logger.debug(
        "Downloading/building wheel for {} into cache_dir {}".format(package, cache_dir)
    )
    abs_cache_dir = os.path.abspath(os.path.expanduser(cache_dir))
    cwd_cache_dir = abs_cache_dir.replace(os.getcwd(), ".", 1)
    args = _get_wheel_args(index_url, extra_index_url, pre, cache_dir) + [package]
    try:
        out = stream_bash_command(args)
    except subprocess.CalledProcessError as err:
        output = getattr(err, "output") or ""
        logger.error(
            "Downloading/building wheel for {} failed with output:\n{}".format(
                package, output.strip()
            )
        )
        raise RuntimeError("Failed to download/build wheel for {}".format(package))
    out = out.splitlines()[::-1]
    abs_cache_dir_lower = abs_cache_dir.lower()
    cwd_cache_dir_lower = cwd_cache_dir.lower()
    cache_dir_lower = cache_dir.lower()
    for i, line in enumerate(out):
        line = line.strip()
        line_lower = line.lower()
        if not (
            cache_dir_lower in line_lower
            or abs_cache_dir_lower in line_lower
            or cwd_cache_dir_lower in line_lower
            or "stored in directory" in line_lower
        ):
            continue

        if "stored in directory" in line_lower:
            # wheel was built
            fnames = [
                part.replace("filename=", "")
                for part in out[i + 1].split()
                if part.startswith("filename=")
            ]
            if not fnames:
                # older pip might not state filename, only directory
                whldir = line.replace("Stored in directory:", "").strip()
                dirpath, _, filenames = next(os.walk(whldir))
                all_wheels = sorted(
                    (f for f in filenames if f.endswith(".whl")),
                    key=lambda x: os.path.getmtime(os.path.join(dirpath, x)),
                    reverse=True,
                )
                if package.startswith("."):
                    fname = all_wheels[0]
                else:
                    fname = None
                    pkg = parse_req(package).key
                    for whl in all_wheels:
                        if pkg in canonicalize_name(whl):
                            fname = whl
                            break
                    if not fname:
                        logger.error(
                            "Failed to extract wheel filename from pip stdout: \n{}".format(
                                "\n".join(out[::-1])
                            )
                        )
                        raise RuntimeError(
                            "Failed to find wheel for {} in {}. Wheels found: {}".format(
                                package, whldir, all_wheels
                            )
                        )
            else:
                fname = fnames[0]
        else:
            # wheel was fetched
            # match on lowercase line for windows compatibility
            fname_len = len(
                line_lower.split(
                    abs_cache_dir_lower
                    if abs_cache_dir_lower in line_lower
                    else cwd_cache_dir_lower
                    if cwd_cache_dir_lower in line_lower
                    else cache_dir_lower,
                    1,
                )[1].split(".whl", 1)[0]
                + ".whl"
            )
            fname = line[-fname_len:]

        logger.debug(str({package: os.path.join(cache_dir, fname.lstrip(os.path.sep))}))
        return os.path.join(cache_dir, fname.lstrip(os.path.sep))
    logger.error(
        "Failed to extract wheel filename from pip stdout: \n{}".format(
            "\n".join(out[::-1])
        )
    )
    raise RuntimeError("Failed to download/build wheel for {}".format(package))


def _extract_metadata(wheel_fname):
    wheel_fname = os.path.abspath(wheel_fname)
    logger.debug("Searching metadata in %s", wheel_fname)
    if not os.path.exists(wheel_fname):
        raise RuntimeError("File not found: {}".format(wheel_fname))
    info = get_metadata(wheel_fname)
    if info is None:
        raise RuntimeError("Failed to extract metadata: {}".format(wheel_fname))
    data = vars(info)
    data.pop("filename", None)
    return data


def _get_wheel_requirements(metadata, extras_requested):
    """Extract the immediate dependencies from wheel metadata."""
    all_requires = metadata.get("requires_dist", [])
    if not all_requires:
        return []
    result = []
    env_data = default_environment()
    for req_str in all_requires:
        req = parse_req(req_str)
        req_short, _sep, _marker = str(req).partition(";")
        if req.marker is None:
            # unconditional dependency
            result.append(req_short)
            continue
        # conditional dependency - must be evaluated in environment context
        for extra in [None] + extras_requested:
            if req.marker.evaluate(dict(env_data, extra=extra)):
                logger.debug("included conditional dep %s", req_str)
                result.append(req_short)
                break
        else:
            logger.debug("dropped conditional dep %s", req_str)
    result = sorted(set(result))  # this makes the dep tree deterministic/repeatable
    return result


def is_unneeded_dep(package):
    """Evaluate a single package in the context of the current environment."""
    return not _get_wheel_requirements({"requires_dist": [package]}, [])


def discover_dependencies_and_versions(
    package, index_url, extra_index_url, cache_dir, pre
):
    """Get information for a package.

    Args:
        package (str): pip requirement format spec compliant package
        index_url (str): primary PyPI index url
        extra_index_url (str): secondary PyPI index url
        cache_dir (str): directory for storing wheels
        pre (bool): pip --pre flag

    Returns:
        dict: package information:
            'version': the version resolved by pip
            'available': all available versions resolved by pip
            'requires': all requirements as found in corresponding wheel (dist_requires)

    """
    req = parse_req(package)

    extras_requested = sorted(req.extras)

    logger.info("discovering %s", req)
    wheel_fname = _download_wheel(
        req.__str__(), index_url, extra_index_url, pre, cache_dir
    )
    wheel_metadata = _extract_metadata(wheel_fname)
    wheel_requirements = _get_wheel_requirements(wheel_metadata, extras_requested)
    wheel_version = req.url or wheel_metadata["version"]
    available_versions = (
        _get_available_versions(req.extras_name, index_url, extra_index_url, pre)
        if req.key != "." and req.url is None
        else [wheel_version]
    )
    if wheel_version not in available_versions:
        available_versions.append(wheel_version)

    return {
        "name": wheel_metadata["name"],
        "version": wheel_version,
        "available": available_versions,
        "requires": wheel_requirements,
    }



from typing import Dict, Generator, Hashable, List, Optional

from pipgrip.libs.mixology.incompatibility_cause import (
    ConflictCause,
    DependencyCause,
    IncompatibilityCause,
    NoVersionsCause,
    PackageNotFoundCause,
    RootCause,
)
from pipgrip.libs.mixology.package import Package
from pipgrip.libs.mixology.term import Term


class Incompatibility:
    def __init__(
        self, terms, cause
    ):  # type: (List[Term], IncompatibilityCause) -> None
        # Remove the root package from generated incompatibilities, since it will
        # always be satisfied. This makes error reporting clearer, and may also
        # make solving more efficient.
        if (
            len(terms) != 1
            and isinstance(cause, ConflictCause)
            and any(
                term.is_positive() and term.package != Package.root() for term in terms
            )
        ):
            terms = [
                term
                for term in terms
                if not term.is_positive() or term.package != Package.root()
            ]

        if (
            len(terms) == 1
            # Short-circuit in the common case of a two-term incompatibility with
            # two different packages (for example, a dependency).
            or len(terms) == 2
            and terms[0].package != terms[-1].package
        ):
            pass
        else:
            # Coalesce multiple terms about the same package if possible.
            by_name = {}  # type: Dict[Hashable, Dict[Hashable, Term]]
            for term in terms:
                new_terms = []

                if term.package not in by_name:
                    by_name[term.package] = {}

                by_ref = by_name[term.package]
                ref = term.package

                if ref in by_ref:
                    intersection = by_ref[ref].intersect(term)
                    if intersection is not None:
                        by_ref[ref] = intersection
                    else:
                        # package requires itself
                        new_terms.append(by_ref[ref])
                        by_ref[ref] = term

                    # If we have two terms that refer to the same package but have a null
                    # intersection, they're mutually exclusive, making this incompatibility
                    # irrelevant, since we already know that mutually exclusive version
                    # ranges are incompatible. We should never derive an irrelevant
                    # incompatibility.
                    assert by_ref[ref] is not None
                else:
                    by_ref[ref] = term

                for by_ref in by_name.values():
                    positive_terms = [
                        term for term in by_ref.values() if term.is_positive()
                    ]
                    if positive_terms:
                        new_terms += positive_terms
                        continue

                    new_terms += list(by_ref.values())

                terms = new_terms

        self._terms = terms
        self._cause = cause

    @property
    def terms(self):  # type: () -> List[Term]
        return self._terms

    @property
    def cause(self):  # type: () -> IncompatibilityCause
        return self._cause

    @property
    def external_incompatibilities(self):  # type: () -> Generator[Incompatibility]
        """
        Returns all external incompatibilities in this incompatibility's
        derivation graph.
        """
        if isinstance(self._cause, ConflictCause):
            cause = self._cause  # type: ConflictCause
            for incompatibility in cause.conflict.external_incompatibilities:
                yield incompatibility

            for incompatibility in cause.other.external_incompatibilities:
                yield incompatibility
        else:
            yield self

    def is_failure(self):  # type: () -> bool
        return len(self._terms) == 0 or (
            len(self._terms) == 1
            and self._terms[0].package == Package.root()
            and self._terms[0].is_positive()
        )

    def handle_cause(self):  # type: () -> Optional[str]
        if isinstance(self._cause, DependencyCause):
            assert len(self._terms) == 2

            depender = self._terms[0]
            dependee = self._terms[1]
            assert depender.is_positive()
            assert not dependee.is_positive()

            return "{} depends on {}".format(
                depender.to_string(allow_every=True), dependee.inverse
            )
        elif isinstance(self._cause, NoVersionsCause):
            assert len(self._terms) == 1
            assert self._terms[0].is_positive()

            return "no versions of {} match {}".format(
                self._terms[0].package, self._terms[0].constraint.constraint
            )
        elif isinstance(self._cause, PackageNotFoundCause):
            assert len(self._terms) == 1
            assert self._terms[0].is_positive()

            return "{} doesn't exist".format(self._terms[0].package)
        elif isinstance(self._cause, RootCause):
            assert len(self._terms) == 1
            assert not self._terms[0].is_positive()
            assert self._terms[0].package == Package.root()

            return "{} is {}".format(self._terms[0].package, self._terms[0].constraint)

    def __str__(self):
        cause_string = self.handle_cause()
        if cause_string is not None:
            return cause_string
        elif self.is_failure():
            return "version solving failed"

        if len(self._terms) == 1:
            term = self._terms[0]
            if term.constraint.is_any():
                return "{} is {}".format(
                    term.package, "forbidden" if term.is_positive() else "required"
                )
            else:
                return "{} is {}".format(
                    term.package, "forbidden" if term.is_positive() else "required"
                )

        if len(self._terms) == 2:
            term1 = self._terms[0]
            term2 = self._terms[1]

            if term1.is_positive() == term2.is_positive():
                if term1.is_positive():
                    package1 = (
                        term1.package
                        if term1.constraint.is_any()
                        else self._terse(term1)
                    )
                    package2 = (
                        term2.package
                        if term2.constraint.is_any()
                        else self._terse(term2)
                    )

                    return "{} is incompatible with {}".format(package1, package2)
                else:
                    return "either {} or {}".format(
                        self._terse(term1), self._terse(term2)
                    )

        positive = []
        negative = []

        for term in self._terms:
            if term.is_positive():
                positive.append(term.to_string(allow_every=True))
            else:
                negative.append(term.inverse.to_string())

        if positive and negative:
            if len(positive) == 1:
                positive_term = [term for term in self._terms if term.is_positive()][0]

                return "{} requires {}".format(
                    positive_term.to_string(allow_every=True), " or ".join(negative)
                )
            else:
                return "if {} then {}".format(
                    " and ".join(positive), " or ".join(negative)
                )
        elif positive:
            return "one of {} must be false".format(" or ".join(positive))
        else:
            return "one of {} must be true".format(" or ".join(negative))

    def and_to_string(
        self, other, details, this_line, other_line
    ):  # type: (Incompatibility, dict, int, int) -> str
        requires_both = self._try_requires_both(other, details, this_line, other_line)
        if requires_both is not None:
            return requires_both

        requires_through = self._try_requires_through(
            other, details, this_line, other_line
        )
        if requires_through is not None:
            return requires_through

        requires_forbidden = self._try_requires_forbidden(
            other, details, this_line, other_line
        )
        if requires_forbidden is not None:
            return requires_forbidden

        buffer = [str(self)]
        if this_line is not None:
            buffer.append(" " + this_line)

        buffer.append(" and {}".format(str(other)))

        if other_line is not None:
            buffer.append(" " + other_line)

        return "\n".join(buffer)

    def _try_requires_both(
        self, other, details, this_line, other_line
    ):  # type: (Incompatibility, dict, int, int) -> Optional[str]
        if len(self._terms) == 1 or len(other.terms) == 1:
            return

        this_positive = self._single_term_where(lambda term: term.is_positive())
        if this_positive is None:
            return

        other_positive = other._single_term_where(lambda term: term.is_positive())
        if other_positive is None:
            return

        if this_positive.package != other_positive.package:
            return

        this_negatives = " or ".join(
            [term.inverse.to_string() for term in self._terms if not term.is_positive()]
        )

        other_negatives = " or ".join(
            [term.inverse.to_string() for term in other.terms if not term.is_positive()]
        )

        buffer = [self._terse(this_positive, allow_every=True) + " "]
        is_dependency = isinstance(self.cause, DependencyCause) and isinstance(
            other.cause, DependencyCause
        )

        if is_dependency:
            buffer.append("depends on")
        else:
            buffer.append("requires")

        buffer.append(" both {}".format(this_negatives))
        if this_line is not None:
            buffer.append(" ({})".format(this_line))

        buffer.append(" and {}".format(other_negatives))

        if other_line is not None:
            buffer.append(" ({})".format(other_line))

        return "".join(buffer)

    def _try_requires_through(
        self, other, details, this_line, other_line
    ):  # type: (Incompatibility, dict, int, int) -> Optional[str]
        if len(self._terms) == 1 or len(other.terms) == 1:
            return

        this_negative = self._single_term_where(lambda term: not term.is_positive())
        other_negative = other._single_term_where(lambda term: not term.is_positive())

        if this_negative is None and other_negative is None:
            return

        this_positive = self._single_term_where(lambda term: term.is_positive())
        other_positive = self._single_term_where(lambda term: term.is_positive())

        if (
            this_negative is not None
            and other_positive is not None
            and this_negative.package == other_positive.package
            and this_negative.inverse.satisfies(other_positive)
        ):
            prior = self
            prior_negative = this_negative
            prior_line = this_line
            latter = other
            latter_line = other_line
        elif (
            other_negative is not None
            and this_positive is not None
            and other_negative.package == this_positive.package
            and other_negative.inverse.satisfies(this_positive)
        ):
            prior = other
            prior_negative = other_negative
            prior_line = other_line
            latter = self
            latter_line = this_line
        else:
            return

        prior_positives = [term for term in prior.terms if term.is_positive()]

        buffer = []
        if len(prior_positives) > 1:
            prior_string = " or ".join([self._terse(term) for term in prior_positives])
            buffer.append("if {} then ".format(prior_string))
        else:
            if isinstance(prior.cause, DependencyCause):
                verb = "depends on"
            else:
                verb = "requires"

            buffer.append(
                "{} {} ".format(prior_positives[0].to_string(allow_every=True), verb)
            )

        buffer.append(prior_negative.inverse.to_string())
        if prior_line is not None:
            buffer.append(" ({})".format(prior_line))

        buffer.append(" which ")

        if isinstance(latter.cause, DependencyCause):
            buffer.append("depends on ")
        else:
            buffer.append("requires ")

        buffer.append(
            " or ".join(
                [
                    term.inverse.to_string()
                    for term in latter.terms
                    if not term.is_positive()
                ]
            )
        )

        if latter_line is not None:
            buffer.append(" ({})".format(latter_line))

        return "".join(buffer)

    def _try_requires_forbidden(
        self, other, details, this_line, other_line
    ):  # type: (Incompatibility, dict, int, int) -> Optional[str]
        if len(self._terms) != 1 and len(other.terms) != 1:
            return None

        if len(self.terms) == 1:
            prior = other
            latter = self
            prior_line = other_line
            latter_line = this_line
        else:
            prior = self
            latter = other
            prior_line = this_line
            latter_line = other_line

        negative = prior._single_term_where(lambda term: not term.is_positive())
        if negative is None:
            return

        if not negative.inverse.satisfies(latter.terms[0]):
            return

        positives = [t for t in prior.terms if t.is_positive()]

        buffer = []
        if len(positives) > 1:
            prior_string = " or ".join([self._terse(term) for term in positives])
            buffer.append("if {} then ".format(prior_string))
        else:
            buffer.append(self._terse(positives[0], allow_every=True))
            if isinstance(prior.cause, DependencyCause):
                buffer.append(" depends on ")
            else:
                buffer.append(" requires ")

        buffer.append(latter.terms[0].to_string(allow_every=True) + " ")
        if prior_line is not None:
            buffer.append("({}) ".format(prior_line))

        if isinstance(latter.cause, NoVersionsCause):
            buffer.append("which doesn't match any versions")
        elif isinstance(latter.cause, PackageNotFoundCause):
            buffer.append("which doesn't exist")
        else:
            buffer.append("which is forbidden")

        if latter_line is not None:
            buffer.append(" ({})".format(latter_line))

        return "".join(buffer)

    def _terse(self, term, allow_every=False):  # type: (Term, bool) -> str
        return term.to_string(allow_every=allow_every)

    def _single_term_where(self, callable_):  # type: (callable) -> Optional[Term]
        found = None
        for term in self._terms:
            if not callable_(term):
                continue

            if found is not None:
                return

            found = term

        return found

    def __repr__(self):
        return "<{} {}>".format(self.__class__.__name__, str(self))



import io
import os.path as path

from setuptools import find_packages, setup

here = path.abspath(path.dirname(__file__))

requirements_path = path.join(here, "requirements", "prod.txt")

readme_path = path.join(here, "README.md")


def read_requirements(path):
    try:
        with io.open(path, mode="rt", encoding="utf-8") as fp:
            return list(filter(None, (line.split("#")[0].strip() for line in fp)))
    except IndexError:
        raise RuntimeError("{} is broken".format(path))


def read_readme(path):
    with io.open(path, mode="rt", encoding="utf-8") as fp:
        return fp.read()


setup(
    name="pipgrip",
    description="Lightweight pip dependency resolver with deptree preview functionality based on the PubGrub algorithm",
    long_description=read_readme(readme_path),
    long_description_content_type="text/markdown",
    setup_requires=[
        "setuptools_scm<6; python_version=='2.7'",
        "setuptools_scm<7; python_version>'2.7'",
    ],
    install_requires=read_requirements(requirements_path),
    use_scm_version={
        "version_scheme": "guess-next-dev",
        "local_scheme": "dirty-tag",
        "write_to": "src/pipgrip/_repo_version.py",
        "write_to_template": 'version = "{version}"\n',
        "relative_to": __file__,
    },
    include_package_data=True,
    package_data={},
    packages=find_packages(where="src"),
    package_dir={"": "src"},
    entry_points={"console_scripts": ["pipgrip = pipgrip.cli:main"]},
    author="ddelange",
    author_email="ddelange@delange.dev",
    url="https://github.com/ddelange/pipgrip",
    python_requires=">=2.7, !=3.0.*, !=3.1.*, !=3.2.*, !=3.3.*, !=3.4.*",
    classifiers=[
        "Development Status :: 5 - Production/Stable",
        "Environment :: Console",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: BSD License",
        "Operating System :: OS Independent",
        "Programming Language :: Python",
        "Programming Language :: Python :: 2",
        "Programming Language :: Python :: 2.7",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.5",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Programming Language :: Python :: 3.11",
        "Topic :: Software Development :: Libraries :: Python Modules",
        "Topic :: System :: Archiving :: Packaging",
        "Topic :: System :: Systems Administration",
        "Topic :: System :: Installation/Setup",
        "Topic :: Utilities",
    ],
    keywords="pure-Python pip exhaustive package dependency version conflict resolver PyPI PubGrub pipdeptree",
    license="BSD-3-Clause",
)
